"""
Tools for working with volumes of interest (VOIs).

A volume of interest is a sub-volume contained within a full dataset.
It is represented by the [`VOI`][hoa_tools.voi.VOI] class.

All coordinates are given in 'array coordinates',
where 0 is the centre of the first voxel, 1 is the centre of the second voxel etc.
"""

import itertools
from math import ceil, floor
from typing import Any

import SimpleITK as sitk
import xarray as xr
from pydantic import BaseModel

from hoa_tools.dataset import Dataset
from hoa_tools.registration import Inventory as RegInventory
from hoa_tools.types import ArrayCoordinate


class VOI(BaseModel):
    """
    A volume of interest attached to a given dataset.
    """

    dataset: Dataset
    """Dataset that this VOI is in."""
    downsample_level: int
    """Downsampling level of the dataset that this VOI is defined in."""
    lower_corner: ArrayCoordinate
    """Index of lower corner in array coordinates."""
    size: ArrayCoordinate
    """Size of VOI in array coordinates."""

    @property
    def voxel_size_um(self) -> float:
        """
        Voxel size in micrometers.
        """
        return float(self.dataset.data.voxel_size_um * 2 ** (self.downsample_level))

    @property
    def upper_corner(self) -> ArrayCoordinate:
        """
        Upper corner of the VOI.
        """
        return ArrayCoordinate(
            x=self.lower_corner.x + self.size.x,
            y=self.lower_corner.y + self.size.y,
            z=self.lower_corner.z + self.size.z,
        )

    @property
    def corners(self) -> list[ArrayCoordinate]:
        """
        All 8 corners of the VOI.
        """
        corner_tuples: list[tuple[int, int, int]] = list(
            itertools.product(  # type: ignore[arg-type]
                *zip(
                    [self.lower_corner.x, self.lower_corner.y, self.lower_corner.z],
                    [self.upper_corner.x, self.upper_corner.y, self.upper_corner.z],
                    strict=True,
                )
            )
        )
        return [ArrayCoordinate(x=i[0], y=i[1], z=i[2]) for i in corner_tuples]

    def get_data_array(self) -> xr.DataArray:
        """
        Get data array for this VOI.
        """
        da = self.dataset.data_array(downsample_level=self.downsample_level)
        return da.isel(
            x=slice(self.lower_corner.x, self.upper_corner.x),
            y=slice(self.lower_corner.y, self.upper_corner.y),
            z=slice(self.lower_corner.z, self.upper_corner.z),
        )

    def get_sitk_image(self) -> sitk.Image:
        """
        Get a SimpleITK image of this VOI.
        """
        data_array = self.get_data_array()
        image = sitk.GetImageFromArray(data_array.transpose("x", "y", "z").values)
        image.SetSpacing((self.voxel_size_um, self.voxel_size_um, self.voxel_size_um))  # type: ignore[no-untyped-call]
        image.SetOrigin(  # type: ignore[no-untyped-call]
            (
                float(data_array.coords["z"][0]),
                float(data_array.coords["y"][0]),
                float(data_array.coords["x"][0]),
            )
        )
        return image

    def get_data_array_on_voi(
        self,
        target_voi: "VOI",
        *,
        interpolator: Any = sitk.sitkLinear,
        transform: sitk.Transform | None = None,
    ) -> xr.DataArray:
        """
        Get data array for this VOI resampled to the grid of another VOI.

        If the transform isn't given, uses the transform between two datasets in the
        registration inventory.

        Parameters
        ----------
        target_voi :
            VOI to resample data in this VOI on to.
        interpolator :
            Interpolation method to use.
            See https://simpleitk.org/doxygen/v2_4/html/namespaceitk_1_1simple.html#a7cb1ef8bd02c669c02ea2f9f5aa374e5
        transform :
            If given, transform used to map this VOI on to the target VOI.
            If not given, transform is taken from the registration inventory.

        """
        if transform is None:
            transform = RegInventory.get_registration(
                source_dataset=self.dataset, target_dataset=target_voi.dataset
            )
        default_value = 0
        new_image = sitk.Resample(
            self.get_sitk_image(),
            target_voi.get_sitk_image(),
            transform.GetInverse(),  # type: ignore[no-untyped-call]
            interpolator,
            default_value,
        )

        new_array = target_voi.get_data_array().copy()
        new_array.values = sitk.GetArrayFromImage(new_image).T
        return new_array

    def change_downsample_level(self, *, new_downsample_level: int) -> "VOI":
        """
        Return a new VOI at a different downsample level.
        """
        resolution_ratio = (2**self.downsample_level) / (2**new_downsample_level)
        new_lower_corner = ArrayCoordinate(
            x=floor(self.lower_corner.x * resolution_ratio),
            y=floor(self.lower_corner.y * resolution_ratio),
            z=floor(self.lower_corner.z * resolution_ratio),
        )
        new_size = ArrayCoordinate(
            x=ceil(self.size.x * resolution_ratio),
            y=ceil(self.size.y * resolution_ratio),
            z=ceil(self.size.z * resolution_ratio),
        )
        return VOI(
            dataset=self.dataset,
            downsample_level=new_downsample_level,
            lower_corner=new_lower_corner,
            size=new_size,
        )

    def transform_to(
        self,
        dataset: Dataset,
        *,
        transform: sitk.Transform | None = None,
    ) -> "VOI":
        """
        Transform this VOI to another dataset.

        The new VOI will completely contain the transformed original VOI.

        Parameters
        ----------
        dataset :
            Dataset to transform to.
        transform :
            If given, transform used to map this VOI on to the target VOI.
            If not given, transform is taken from the registration inventory.

        """
        if transform is None:
            if (self.dataset, dataset) not in RegInventory:
                msg = (
                    f"Transform between {self.dataset.name} and {dataset.name} "
                    "not found in registration inventory."
                )
                raise RuntimeError(msg)

            transform = RegInventory.get_registration(
                source_dataset=self.dataset, target_dataset=dataset
            )

        old_voi = self.change_downsample_level(new_downsample_level=0)

        # Convert to physical space
        physical_corners = [
            c.to_physical_coordinate(voxel_size=self.dataset.data.voxel_size_um)
            for c in old_voi.corners
        ]
        # Transform
        physical_corners_transformed = [
            c.transform(transform) for c in physical_corners
        ]
        # Convert back to array space
        corners_transformed = [
            c.to_array_coordinate(voxel_size=dataset.data.voxel_size_um)
            for c in physical_corners_transformed
        ]
        lower_corner = ArrayCoordinate(
            z=min([c.z for c in corners_transformed]),
            y=min([c.y for c in corners_transformed]),
            x=min([c.x for c in corners_transformed]),
        )
        upper_corner = ArrayCoordinate(
            z=max([c.z for c in corners_transformed]) + 1,
            y=max([c.y for c in corners_transformed]) + 1,
            x=max([c.x for c in corners_transformed]) + 1,
        )
        return VOI(
            dataset=dataset,
            downsample_level=0,
            lower_corner=lower_corner,
            size=ArrayCoordinate(
                x=upper_corner.x - lower_corner.x,
                y=upper_corner.y - lower_corner.y,
                z=upper_corner.z - lower_corner.z,
            ),
        )



"""
Tools for working with the dataset inventory.
"""

import pandas as pd

from hoa_tools.dataset import _DATASETS


def load_inventory() -> pd.DataFrame:
    """
    Load the dataset inventory.

    Returns
    -------
    inventory :
        Dataset inventory.

    """
    data = [
        {
            "donor": d.donor.id,
            "organ": d.sample.organ,
            "organ_context": d.sample.organ_context,
            "voi": d.voi,
            "voxel_size_um": d.data.voxel_size_um,
        }
        for d in _DATASETS.values()
    ]
    return pd.DataFrame(data=data, index=list(_DATASETS.keys()))



import hoa_tools.dataset
import hoa_tools.registration
import hoa_tools.voi
from hoa_tools.types import ArrayCoordinate


def test_transform_voi() -> None:
    overview = hoa_tools.dataset.get_dataset(
        "S-20-29_brain_complete-organ_25.33um_bm05"
    )
    child = hoa_tools.dataset.get_dataset("S-20-29_brain_VOI-04_6.5um_bm05")

    child_voi = hoa_tools.voi.VOI(
        dataset=child,
        downsample_level=0,
        lower_corner=ArrayCoordinate(x=3434, y=2060, z=2656),
        size=ArrayCoordinate(x=256, y=256, z=128),
    )

    overview_voi = child_voi.transform_to(overview)
    assert overview_voi.dataset.name == "S-20-29_brain_complete-organ_25.33um_bm05"
    assert overview_voi.downsample_level == 0
    assert overview_voi.lower_corner == ArrayCoordinate(x=2975, y=1689, z=4316)
    assert overview_voi.size == ArrayCoordinate(x=69, y=69, z=33)



# # Fetching data

# In this tutorial we will fetch some of the data. All public Human Organ Atlas
# data is available through a read-only public Google Cloud Storage bucket.
# in a chunked data format, making it easy to access sub-volumes of data.

# ## Getting a remote data store

# First we'll fetch a `Dataset` object from the inventory.

import hoa_tools.dataset
import hoa_tools.inventory

inventory = hoa_tools.inventory.load_inventory()
dataset_name = inventory.index[0]
print(dataset_name)

dataset = hoa_tools.dataset.get_dataset(dataset_name)
print(dataset)

# The `Dataset.remote_array` property gives us access to a remote data array for a dataset.
# The ``level`` parameter lets us choose a downsampling level: ``level=0`` is the original
# resolution dataset, and each time the level goes up by one the dataset is downsampled
# by a factor of 2. All datasets are downsampled to ``level=4``, so lets get a remote store
# for the lowest resolution copy of the data.

data_array = dataset.data_array(downsample_level=4)
data_array

# Here we can see that the array is read-only, 16 bit, and has shape ``(477, 238, 238)``.
# At this point no data has been downloaded - to download data you need to index the remote array.
# As an example, lets fetch and show a slice in the x-y plane::


middle_slice = data_array.isel(z=437)
im = middle_slice.plot(cmap="Grays_r")
im.axes.set_aspect("equal")


# If you want to save a part of the array for later analysis, see the [xarray guide
# to reading and writing files](https://docs.xarray.dev/en/stable/user-guide/io.html).



# # Dataset Inventory

# In the first tutorial we will show how to search for available data
# using the `hoa_tools.inventory` module. By the end you should know
# how to search and filter the built in dataset inventory.

import pandas as pd

import hoa_tools.inventory

pd.options.display.width = None  # type: ignore[assignment]

# ## Loading the inventory
#
# First we'll load the inventory::


inventory = hoa_tools.inventory.load_inventory()
inventory

# The inventory is a `pandas.DataFrame` object. Each row is a different dataset, and each column
# contains the properties of the dataset. The available columns are::

print(list(inventory.columns))

# ## Searching the inventory
#
# To find all the spleen datasets we can filter the dataframe::

inventory[inventory["organ"] == "spleen"]



from hoa_tools.dataset import get_dataset
from hoa_tools.voi import VOI


def test_voi_properties() -> None:
    dataset = get_dataset("LADAF-2020-27_spleen_complete-organ_25.08um_bm05")
    voi = VOI(
        dataset=dataset,
        downsample_level=2,
        lower_corner={"x": 1, "y": 2, "z": 3},
        size={"x": 30, "y": 20, "z": 10},
    )

    assert voi.voxel_size_um == 100.32



"""
Common types used across the library.
"""

from math import floor
from typing import Self

import SimpleITK as sitk
from pydantic import BaseModel


class PhysicalCoordinate(BaseModel):
    """
    A single coordinate in physical space.
    """

    x: float
    y: float
    z: float

    def to_array_coordinate(self, *, voxel_size: float) -> "ArrayCoordinate":
        """
        Given a voxel size, convert this physical coordinate to an array coordinate.
        """
        return ArrayCoordinate(
            x=floor(self.x / voxel_size),
            y=floor(self.y / voxel_size),
            z=floor(self.z / voxel_size),
        )

    def transform(self, t: sitk.Transform) -> Self:
        """
        Transform this coordinate.

        Notes
        -----
        The transform is applied in zyx order.

        """
        new_coord = t.TransformPoint((self.z, self.y, self.x))  # type: ignore[no-untyped-call]
        return self.__class__(z=new_coord[0], y=new_coord[1], x=new_coord[2])


class ArrayCoordinate(BaseModel):
    """
    A single coordinate in array space.
    """

    x: int
    y: int
    z: int

    def to_physical_coordinate(self, *, voxel_size: float) -> "PhysicalCoordinate":
        """
        Given a voxel size, convert this array coordinate to a physical coordinate.
        """
        return PhysicalCoordinate(
            x=self.x * voxel_size, y=self.y * voxel_size, z=self.z * voxel_size
        )



# # Individual datasets
#
# In this chapter we'll look at what we can do with an individual dataset.
# In ``hoa-tools`` individual datasets are represented by `Dataset` objects.

import pandas as pd

import hoa_tools.dataset
import hoa_tools.inventory

pd.options.display.width = None  # type: ignore[assignment]

# First, lets print all the spleen datasets available from the inventory::

inventory = hoa_tools.inventory.load_inventory()
spleen_inventory = inventory[inventory["organ"] == "spleen"]
spleen_inventory

# We can see that three datasets are available. They are from the same organ, with one being the full
# organ dataset and two being high resolution region of interest datasets. Lets get the full organ
# dataset

whole_spleen = hoa_tools.dataset.get_dataset(
    "LADAF-2020-27_spleen_complete-organ_25.08um_bm05"
)
whole_spleen

# Because this is a full-organ dataset, it will have a number of child datasets. These are scans of
# the same organ taken at a higher resolution over a subset of the full volume. For our selected
# dataset these child datasets are::

child_datasets = whole_spleen.get_children()
for dataset in child_datasets:
    print(dataset)

# We can see there are two child datasets, one at a resolution of 1.29 μm and one at a
# resolution of 6.05 μm.



"""
Tools for working with individual datasets.

The core class representing a single dataset is [`Dataset`][hoa_tools.dataset.Dataset].
This inherits from [`HOAMetadata`][hoa_tools.metadata.HOAMetadata],
which stores all the metadata for a given dataset.

[`Dataset`][hoa_tools.dataset.Dataset] objects are not designed to be created by users.
To get a [`Dataset`][hoa_tools.dataset.Dataset], use the
[`get_dataset`][hoa_tools.dataset.get_dataset] function in this module.
"""

from functools import cached_property
from pathlib import Path
from typing import Literal

import dask.array.core
import gcsfs
import numpy as np
import xarray as xr
import zarr.core
import zarr.n5
import zarr.storage

from hoa_tools._n5 import N5FSStore
from hoa_tools.metadata import HOAMetadata

__all__ = ["Dataset", "get_dataset"]


class Dataset(HOAMetadata):
    """
    An individual Human Organ Atlas dataset.
    """

    def __str__(self) -> str:
        return f"Dataset(name={self.name})"

    def _organ_str(self) -> str:
        """
        Get name of organ, with organ context appended if present.
        """
        organ_str = str(self.sample.organ)
        if self.sample.organ_context:
            organ_str += "_" + self.sample.organ_context
        return organ_str

    def _voxel_size_str(self) -> str:
        # Make sure that voxel size includes a .0 if an integer value.
        return str(float(self.data.voxel_size_um))

    @property
    def is_full_organ(self) -> bool:
        """
        Whether this dataset contains the whole organ or not.
        """
        return self.voi.startswith("complete")

    @property
    def is_zoom(self) -> bool:
        """
        Whether this dataset is a zoom or not.
        """
        return not self.is_full_organ

    def get_children(self) -> list["Dataset"]:
        """
        Get child dataset(s).

        Child datasets are high-resolution zooms of a full-organ dataset.

        Notes
        -----
        For full-organ datasets, this returns an empty list.

        """
        children = [
            d
            for d in _DATASETS.values()
            if (
                d.donor.id == self.donor.id
                and d.sample.organ == self.sample.organ
                and d.scan.beamline == self.scan.beamline
                and d.is_zoom
            )
        ]
        return sorted(children, key=lambda c: c.name)

    def get_parents(self) -> list["Dataset"]:
        """
        Get parent dataset(s).

        Parent datasets are full-organ datasets from which a zoom dataset is taken from.

        Notes
        -----
        For zoom datasets, this returns an empty list.

        """
        parents = [
            d
            for d in _DATASETS.values()
            if (
                d.donor.id == self.donor.id
                and d.sample.organ == self.sample.organ
                and d.scan.beamline == self.scan.beamline
                and d.is_full_organ
            )
        ]
        return sorted(parents, key=lambda c: c.name)

    @property
    def _remote_fmt(self) -> Literal["n5", "zarr"]:
        if self.data.gcs_url.startswith("n5://"):
            return "n5"
        if self.data.gcs_url.startswith("zarr://"):
            return "zarr"
        raise RuntimeError("URL must start with n5:// or zarr://")

    @cached_property
    def _remote_store(self) -> zarr.Group:
        """
        Remote data store.
        """
        gcs_url = self.data.gcs_url
        gcs_path = gcs_url.removeprefix("n5://gs://").removeprefix("zarr://gs://")

        # n5://gs://ucl-hip-ct-35a68e99feaae8932b1d44da0358940b/S-20-29/heart/2.5um_VOI-01_bm05/
        bucket, path = gcs_path.split("/", maxsplit=1)
        fs = gcsfs.GCSFileSystem(project=bucket, token="anon", access="read_only")  # noqa: S106
        if self._remote_fmt == "n5":
            store = N5FSStore(url=bucket, fs=fs, mode="r")
        elif self._remote_fmt == "zarr":
            store = zarr.storage.FSStore(url=bucket, fs=fs, mode="r")

        return zarr.open_group(store, mode="r", path=path)

    def _remote_array(self, *, downsample_level: int) -> zarr.core.Array:
        """
        Get an object representing the data array in the remote Google Cloud Store.
        """
        if not downsample_level >= 0:
            msg = "level must be >= 0"
            raise ValueError(msg)

        if self._remote_fmt == "n5":
            key = f"s{downsample_level}"
        else:
            key = f"{downsample_level}"
        return self._remote_store[key]

    def data_array(self, *, downsample_level: int) -> xr.DataArray:
        """
        Get a DataArray representing the array for this image.
        """
        remote_array = self._remote_array(downsample_level=downsample_level)
        dask_array = dask.array.core.from_array(  # type: ignore[no-untyped-call]
            remote_array, chunks=remote_array.chunks
        )
        if self._remote_fmt == "zarr":
            dask_array = dask_array.T

        spacing = self.data.voxel_size_um * 2**downsample_level
        return xr.DataArray(
            dask_array,
            name=self.name,
            dims=["z", "y", "x"],
            coords={
                "z": xr.DataArray(
                    data=(np.arange(dask_array.shape[0]) * spacing),
                    dims=["z"],
                    attrs={"units": "μm"},
                ),
                "y": xr.DataArray(
                    data=(np.arange(dask_array.shape[1]) * spacing),
                    dims=["y"],
                    attrs={"units": "μm"},
                ),
                "x": xr.DataArray(
                    data=(np.arange(dask_array.shape[2]) * spacing),
                    dims=["x"],
                    attrs={"units": "μm"},
                ),
            },
        )


def _get_datasets(data_dir: Path) -> dict[str, Dataset]:
    """
    Load dataset metadtata files.
    """
    return {
        f.stem: Dataset.model_validate_json(f.read_text())
        for f in (data_dir).glob("*.json")
    }


_META_DIR = Path(__file__).parent / "data" / "metadata" / "metadata"
_DATASETS = _get_datasets(_META_DIR)
if len(_DATASETS) == 0:
    raise ImportError(
        "Did not find any dataset metadata files. "
        "This means there is something wrong with the hoa-tools installation. "
        "If you are installing from source, you might need to "
        "initialise and fetch git submodules. "
    )


def get_dataset(name: str) -> Dataset:
    """
    Get a dataset from its name.

    The name of datasets can be looked up using the `hoa_tools.inventory` module.
    """
    return _DATASETS[name]


def change_metadata_directory(data_dir: Path) -> None:
    """
    Update available datasets from another directory of metadata files.

    Designed for internal project members to load metadata files that aren't yet public.
    """
    global _DATASETS  # noqa: PLW0603
    _DATASETS = _get_datasets(data_dir)
    if len(_DATASETS) == 0:
        raise FileNotFoundError(
            f"Did not find any dataset metadata files at {data_dir}"  # noqa: EM102
        )



# generated by datamodel-codegen:
#   filename:  metadata-schema.json
#   timestamp: 2025-04-10T13:29:59+00:00

from __future__ import annotations

from datetime import date
from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field, RootModel


class Orcid(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="ORCID iD",
            pattern="^\\d{4}-\\d{4}-\\d{4}-\\d{3}[0-9X]{1}$",
            title="Orcid",
        ),
    ]


class Contributor(BaseModel):
    first_name: Annotated[str, Field(title="First Name")]
    last_name: Annotated[str, Field(title="Last Name")]
    orcid: Annotated[Orcid | None, Field(description="ORCID iD", title="Orcid")] = None
    roles: Annotated[
        list[
            Literal[
                "Conceptualization",
                "Data curation",
                "Formal analysis",
                "Funding acquisition",
                "Investigation",
                "Methodology",
                "Project administration",
                "Resources",
                "Software",
                "Supervision",
                "Validation",
                "Vlisualization",
                "Writing - original draft",
                "Writing - review & editing",
            ]
        ],
        Field(
            description="Contributor role(s), following the Contributor Role Taxonomy (CReditT)",
            title="Roles",
        ),
    ]


class Data(BaseModel):
    shape: Annotated[
        list,
        Field(
            description="Array shape for dataset.",
            max_length=3,
            min_length=3,
            title="Shape",
        ),
    ]
    voxel_size_um: Annotated[
        float,
        Field(
            description="Isotropic size of a single voxel, in micrometers.",
            gt=0.0,
            title="Voxel size",
        ),
    ]
    gcs_url: Annotated[
        str,
        Field(
            description="URL to data in Google Cloud Storage. Starts with either n5:// or zarr:// to specify whether data is in N5 or OME-Zarr 0.4 format.",
            title="Google cloud storage URL",
        ),
    ]


class Age(RootModel[int]):
    root: Annotated[int, Field(description="Age in years at death.", ge=0, title="Age")]


class Weight(RootModel[float]):
    root: Annotated[
        float, Field(description="Weight in kg at death.", gt=0.0, title="Weight")
    ]


class Height(RootModel[float]):
    root: Annotated[
        float, Field(description="Height in cm at death.", gt=0.0, title="Height")
    ]


class Donor(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    id: Annotated[str, Field(description="Unique donor ID.", title="ID")]
    age: Annotated[
        Age | Literal["<18", ">18"] | None,
        Field(description="Age in years at death.", title="Age"),
    ] = None
    sex: Annotated[Literal["M", "F"] | None, Field(title="Sex")] = None
    weight: Annotated[
        Weight | None, Field(description="Weight in kg at death.", title="Weight")
    ] = None
    height: Annotated[
        Height | None, Field(description="Height in cm at death.", title="Height")
    ] = None
    cause_of_death: Annotated[str | None, Field(title="Cause of death")] = None
    date_of_death: Annotated[date | None, Field(title="Date of death")] = None
    medical_history: Annotated[str | None, Field(title="Medical history")] = None
    diabetes: Annotated[
        Literal["Type 2", "No", "Yes"] | None, Field(title="Diabetes history")
    ] = None
    hypertension: Annotated[
        Literal["Yes", "No"] | None, Field(title="Hypertension history")
    ] = None
    smoker: Annotated[
        Literal["Never", "Yes"] | None, Field(title="Smoking history")
    ] = None


class Proposal(BaseModel):
    proposal_number: Annotated[
        Literal["md1252", "md1290", "md1389"],
        Field(description="Proposal number.", title="Proposal Number"),
    ]
    title: Annotated[str, Field(description="Proposal title.", title="Title")]
    proposers: Annotated[
        list[str], Field(description="Proposal proposers.", title="Proposers")
    ]


class Registration(BaseModel):
    source_dataset: Annotated[
        str,
        Field(
            description="Name of dataset that is registered (this dataset).",
            title="Source Dataset",
        ),
    ]
    target_dataset: Annotated[
        str,
        Field(
            description="Name of dataset that this dataset is registered to.",
            title="Target Dataset",
        ),
    ]
    translation: Annotated[
        list,
        Field(
            description="Translation vector.",
            max_length=3,
            min_length=3,
            title="Translation",
        ),
    ]
    rotation: Annotated[
        float, Field(description="Rotation in degrees about z-axis", title="Rotation")
    ]
    scale: Annotated[
        float, Field(description="Isotropic scaling factor.", title="Scale")
    ]


class PostMortemIntervalHours(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Interval between death and fixation, in hours.",
            ge=0.0,
            title="Post-mortem interval",
        ),
    ]


class Sample(BaseModel):
    organ: Annotated[
        Literal[
            "brain",
            "breast",
            "colon",
            "eye",
            "femaleGenitalia",
            "kidney",
            "liver",
            "lung",
            "heart",
            "oesophagus",
            "spinalCord",
            "spine",
            "spleen",
            "heart+lungs",
            "pancreas",
            "placenta",
            "prostate",
            "testis",
            "uterus",
            "thoracicBlock",
            "soleus",
            "achillesTendon",
            "knee",
            "tibia",
        ],
        Field(description="Organ name.", title="Organ"),
    ]
    organ_context: Annotated[
        str | None,
        Field(
            description="Human-readable context for location of the dataset within the whole organ. Only present for some zoom datasets.",
            examples=["upper_lobe", "left", "tumor"],
            title="Organ Context",
        ),
    ] = None
    post_mortem_interval_hours: Annotated[
        PostMortemIntervalHours | None,
        Field(
            description="Interval between death and fixation, in hours.",
            title="Post-mortem interval",
        ),
    ] = None
    fixation_method: Annotated[
        Literal["immersion", "perfusion"] | None,
        Field(description="Fixation preparation method.", title="Fixation Method"),
    ] = None
    fixation_medium: Annotated[
        str | None, Field(description="Fixation medium", title="Fixation Medium")
    ] = None
    organ_infilled: Annotated[bool | str | None, Field(title="Organ Infilled")] = None
    stabilisation_medium: Annotated[
        Literal["agar cubes", "crushed agar", "beads"] | None,
        Field(title="Stabilisation Medium"),
    ] = None
    degassing_method: Annotated[
        Literal["inline", "vacuum", "vacuum+inline"] | None,
        Field(title="Degassing Method"),
    ] = None
    scan_solvent: Annotated[
        Literal["formalin", "ethanol"] | None,
        Field(description="Scan medium solvent.", title="Scan Solvent"),
    ] = None
    scan_solvent_concentration: Annotated[
        float | None,
        Field(
            description="Scan medium solvent concentration in percent.",
            title="Solvent concentration",
        ),
    ] = None
    scan_temperature: Annotated[
        str | float | None,
        Field(
            description="Temperature during scan, in degees celsius (if a number).",
            title="Scan Temperature",
        ),
    ] = "room temperature"


class Energy(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Energy used during the scan, measured in keV.",
            gt=0.0,
            title="Energy",
        ),
    ]


class CurrentStart(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Current of the synchrotron at the start of the scan, measured in mA.",
            gt=0.0,
            title="Current Start",
        ),
    ]


class NFrames(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of frames acquired during the scan.",
            gt=0,
            title="Number of frames",
        ),
    ]


class NRef(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of reference images collected during the scan.",
            ge=0,
            title="Number of references",
        ),
    ]


class NDark(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of dark images (without illumination) collected during the scan.",
            ge=0,
            title="Number of darks",
        ),
    ]


class LatencyTime(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Inactive time between active sensor times, measured in seconds.",
            gt=0.0,
            title="Latency Time",
        ),
    ]


class ExposureTime(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Exposure time of a single frame, measured in seconds. This is the sub-frame time multiplied by the number of sub-frames.",
            gt=0.0,
            title="Exposure Time",
        ),
    ]


class SubframeTime(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Exposure time of a single sub-frame, measured in seconds.",
            gt=0.0,
            title="Subframe Time",
        ),
    ]


class NSubframes(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of frames accumulated per exposure.",
            gt=0,
            title="Number of sub-frames",
        ),
    ]


class NScans(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Total number of scans. For a helical scan this is always 1. For a zseries this is the total number of scans concatenated to make the full data volume.",
            gt=0,
            title="Number of scans",
        ),
    ]


class ScanTime(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Total duration of a single scan in seconds.",
            gt=0.0,
            title="Scan Time",
        ),
    ]


class OpticMagnification(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Magnification in the optical setup.",
            gt=0.0,
            title="Optic Magnification",
        ),
    ]


class DistanceSourceSample(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Distance between the source and the sample, measured in millimeters.",
            gt=0.0,
            title="Source-sample distance",
        ),
    ]


class DistanceSampleDetector(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Distance between the sample and the detector, measured in millimeters.",
            gt=0.0,
            title="Sample-detector distance",
        ),
    ]


class SensorRoiXSize(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of pixels in the x-dimension of the sensor's region of interest (ROI).",
            gt=0,
            title="Sensor Roi X Size",
        ),
    ]


class SensorRoiYSize(RootModel[int]):
    root: Annotated[
        int,
        Field(
            description="Number of pixels in the y-dimension of the sensor's region of interest (ROI).",
            gt=0,
            title="Sensor Roi Y Size",
        ),
    ]


class SensorBinning(RootModel[list]):
    root: Annotated[
        list,
        Field(
            description="Pixel binning on the sensor.",
            max_length=2,
            min_length=2,
            title="Sensor Binning",
        ),
    ]


class XrayMagnification(RootModel[float]):
    root: Annotated[
        float,
        Field(
            description="Magnification of X-rays due to divergence between the source and detector.",
            gt=0.0,
            title="Xray Magnification",
        ),
    ]


class Scan(BaseModel):
    date: Annotated[
        date, Field(description="Date when the scan was performed.", title="Date")
    ]
    beamline: Annotated[
        Literal["BM05", "BM18"],
        Field(description="ESRF beamline where scan was performed.", title="Beamline"),
    ]
    energy: Annotated[
        Energy | None,
        Field(
            description="Energy used during the scan, measured in keV.", title="Energy"
        ),
    ] = None
    current_start: Annotated[
        CurrentStart | None,
        Field(
            description="Current of the synchrotron at the start of the scan, measured in mA.",
            title="Current Start",
        ),
    ] = None
    filling_mode: Annotated[
        Literal[
            "16 bunch",
            "7/8 multibunch",
            "uniform multibunch",
            "28*12+1 bunch",
            "4 bunch",
        ]
        | None,
        Field(
            description="Mode of filling used in the synchrotron storage ring.",
            title="Filling Mode",
        ),
    ] = None
    n_frames: Annotated[
        NFrames | None,
        Field(
            description="Number of frames acquired during the scan.",
            title="Number of frames",
        ),
    ] = None
    n_ref: Annotated[
        NRef | None,
        Field(
            description="Number of reference images collected during the scan.",
            title="Number of references",
        ),
    ] = None
    n_dark: Annotated[
        NDark | None,
        Field(
            description="Number of dark images (without illumination) collected during the scan.",
            title="Number of darks",
        ),
    ] = None
    latency_time: Annotated[
        LatencyTime | None,
        Field(
            description="Inactive time between active sensor times, measured in seconds.",
            title="Latency Time",
        ),
    ] = None
    exposure_time: Annotated[
        ExposureTime | None,
        Field(
            description="Exposure time of a single frame, measured in seconds. This is the sub-frame time multiplied by the number of sub-frames.",
            title="Exposure Time",
        ),
    ] = None
    subframe_time: Annotated[
        SubframeTime | None,
        Field(
            description="Exposure time of a single sub-frame, measured in seconds.",
            title="Subframe Time",
        ),
    ] = None
    n_subframes: Annotated[
        NSubframes | None,
        Field(
            description="Number of frames accumulated per exposure.",
            title="Number of sub-frames",
        ),
    ] = None
    scan_type: Annotated[
        Literal["helical", "zseries"],
        Field(description="Type of scan.", title="Scan Type"),
    ]
    scan_range: Annotated[
        float,
        Field(
            description="Angular range of a single scan in degrees.",
            gt=0.0,
            title="Scan Range",
        ),
    ]
    n_scans: Annotated[
        NScans | None,
        Field(
            description="Total number of scans. For a helical scan this is always 1. For a zseries this is the total number of scans concatenated to make the full data volume.",
            title="Number of scans",
        ),
    ] = None
    acquisition: Annotated[
        Literal["half", "quarter"] | None,
        Field(description="Type of tomographic acquisition.", title="Acquisition"),
    ] = None
    z_step: Annotated[
        float | None,
        Field(description="Displacement in millimeters between scans.", title="Z Step"),
    ] = None
    scan_time: Annotated[
        ScanTime | None,
        Field(
            description="Total duration of a single scan in seconds.", title="Scan Time"
        ),
    ] = None
    filters: Annotated[
        list[str] | None,
        Field(description="List of filters used during the scan.", title="Filters"),
    ] = None
    scintillator: Annotated[
        Literal[
            "LuAG:Ce 1000um",
            "LuAG:Ce 100um",
            "LuAG:Ce 100um, reflective",
            "LuAG:Ce 100um, lead glass meniscus",
            "LuAG:Ce 2000um",
            "LuAG:Ce 2000um, reflective",
            "Multi LuAG:Ce 2000um, reflective",
            "LuAG:Ce 200um",
            "LuAG:Ce 250um",
            "LuAG:Ce 250um, reflective",
            "LuAG:Ce 25um",
            "LuAG:Ce 50um",
            "LuAG:Ce 50um, reflective",
            "LuAG:Ce 500um",
            "LySO:Ce 8.5um",
            "LySO:Ce 23um",
            "LySO:Ce 2000um, reflective",
            "Gadox 60um",
            "CSI trixell structured 650um, 5um pitch",
            "GGG 8um",
            "gagg1000_refl",
            "dual-lso5000_refl",
        ]
        | None,
        Field(
            description="Scintillator used for converting X-rays to visible light.",
            title="Scintillator",
        ),
    ] = None
    optic: Annotated[
        Literal[
            "Zoom",
            "DZoom",
            "10x 0.4NA reflective optic",
            "Hasselblad tandem optic 100mm/300mm",
            "Hasselblad revolved 100 100",
            "LAFIP2 optic with canon 50mm",
            "Fixed x2",
            "Fixed x1",
            "Fixed x0.125",
            "Fixed x0.1",
            "Fixed x2.85",
            "Fixed x10",
            "Twinmic 5",
            "Twinmic 10",
            "Twinmic 20",
            "Triplemic 10",
        ]
        | None,
        Field(description="Optical components used during the scan", title="Optic"),
    ] = None
    optic_magnification: Annotated[
        OpticMagnification | None,
        Field(
            description="Magnification in the optical setup.",
            title="Optic Magnification",
        ),
    ] = None
    distance_source_sample: Annotated[
        DistanceSourceSample | None,
        Field(
            description="Distance between the source and the sample, measured in millimeters.",
            title="Source-sample distance",
        ),
    ] = None
    distance_sample_detector: Annotated[
        DistanceSampleDetector | None,
        Field(
            description="Distance between the sample and the detector, measured in millimeters.",
            title="Sample-detector distance",
        ),
    ] = None
    sensor_name: Annotated[
        Literal["PCO edge 4.2 CLHS", "Iris15"] | None,
        Field(
            description="Name of the sensor used during the scan.", title="Sensor Name"
        ),
    ] = None
    sensor_mode: Annotated[
        Literal["rolling shutter", "FFM"] | None,
        Field(
            description="Operating mode of the sensor during the scan.",
            title="Sensor Mode",
        ),
    ] = None
    sensor_roi_x_size: Annotated[
        SensorRoiXSize | None,
        Field(
            description="Number of pixels in the x-dimension of the sensor's region of interest (ROI).",
            title="Sensor Roi X Size",
        ),
    ] = None
    sensor_roi_y_size: Annotated[
        SensorRoiYSize | None,
        Field(
            description="Number of pixels in the y-dimension of the sensor's region of interest (ROI).",
            title="Sensor Roi Y Size",
        ),
    ] = None
    sensor_binning: Annotated[
        SensorBinning | None,
        Field(description="Pixel binning on the sensor.", title="Sensor Binning"),
    ] = None
    pixel_size: Annotated[
        float,
        Field(
            description="Size of a pixel at the sample, measured in micrometers.",
            gt=0.0,
            title="Pixel Size",
        ),
    ]
    xray_magnification: Annotated[
        XrayMagnification | None,
        Field(
            description="Magnification of X-rays due to divergence between the source and detector.",
            title="Xray Magnification",
        ),
    ] = None
    technique: Annotated[
        Literal["Hierarchical Phase-Contrast Tomography (HiP-CT)"],
        Field(title="Experimental technique."),
    ] = "Hierarchical Phase-Contrast Tomography (HiP-CT)"
    experiment_type: Annotated[str | None, Field(title="Experiment Type")] = (
        "Tomography"
    )


class Citation(BaseModel):
    title: Annotated[str, Field(description="Citation title", title="Title")]
    contributors: Annotated[
        list[Contributor],
        Field(
            description="Dataset contributors, listed alphabetically by name.",
            title="Contributors",
        ),
    ]
    doi: Annotated[str | None, Field(description="Dataset DOI", title="Doi")] = None
    author_list: Annotated[
        list[str],
        Field(
            description="Author list. If citing this data, please use this authorship list in the order given",
            title="Author List",
        ),
    ]


class HOAMetadata(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[str, Field(description="Unique name for dataset.", title="Name")]
    dataset_type: Annotated[
        Literal["overview", "zoom"],
        Field(
            description="Type of dataset. Overview datasets contain the full sample. Zoom datasets contain a zoom into a particular volume of interest within the sample.",
            title="Dataset Type",
        ),
    ]
    voi: Annotated[
        str,
        Field(
            description="Volume of interest name. For zoom datasets this gives a unique identifier among zoom datasets taken within the same sample.",
            title="Volume of Interest",
        ),
    ]
    data: Annotated[Data, Field(description="Data metadata")]
    sample: Annotated[Sample, Field(description="Sample metadata")]
    donor: Annotated[Donor, Field(description="Donor metadata")]
    scan: Annotated[Scan, Field(description="Scan metadata")]
    proposal: Annotated[Proposal, Field(description="Beamtime proposal metadata")]
    registration: Annotated[
        Registration | None,
        Field(
            description="Metadata mapping high resolution volume of interest datasets to a full organ dataset. Not present for full organ datasets.",
            title="Registration parameters",
        ),
    ] = None
    citation: Annotated[Citation, Field(title="Dataset citation")]
    metadata_version: Annotated[
        Literal["0.9"],
        Field(description="Version of this metadata file.", title="Metadata Version"),
    ] = "0.9"



"""
Registration inventory and helpers.

A registration defines a transform between two datasets.
The transform is defined in physical space, and in this library data is always
in units of micro-meters (μm).

Transforms are defined using the `SimpleITK` library.
"""

import numpy as np
import SimpleITK as sitk

from hoa_tools.dataset import _DATASETS, Dataset
from hoa_tools.types import PhysicalCoordinate


class RegistrationInventory:
    """
    Inventory of transforms between datasets.

    Transforms are defined as acting on the zyx axes in that order. Note that this
    is different from the order in which the SimpleITK API defines axes in its function
    signatures.
    """

    def __init__(self) -> None:
        """
        Create registration inventory.
        """
        self._registrations: dict[tuple[str, str], sitk.Transform] = {}

    def __contains__(self, item: tuple[Dataset, Dataset]) -> bool:
        """
        Check for existence of registration between two datasets.
        """
        return (item[0].name, item[1].name) in self._registrations

    def get_registration(
        self, *, source_dataset: Dataset, target_dataset: Dataset
    ) -> sitk.Transform:
        """
        Get a registration.
        """
        return self._registrations[(source_dataset.name, target_dataset.name)]

    def add_registration(
        self,
        *,
        source_dataset: Dataset,
        target_dataset: Dataset,
        transform: sitk.Transform,
    ) -> None:
        """
        Add a new transform to the inventory.

        Notes
        -----
        This will override any already defined transforms for these two datasets.

        """
        self._registrations[(source_dataset.name, target_dataset.name)] = transform
        self._registrations[(target_dataset.name, source_dataset.name)] = (
            transform.GetInverse()  # type: ignore[no-untyped-call]
        )


def build_transform(
    *, translation: PhysicalCoordinate, rotation_deg: float, scale: float
) -> sitk.Similarity3DTransform:
    """
    Build a transform from a translation, scale, and rotation.
    """
    axis = (1, 0, 0)
    center = (0, 0, 0)
    return sitk.Similarity3DTransform(  # type: ignore[no-untyped-call]
        scale,
        axis,
        np.deg2rad(rotation_deg),
        (translation.z, translation.y, translation.x),
        center,
    )


Inventory = RegistrationInventory()


def _populate_registrations_from_metadata(inventory: RegistrationInventory) -> None:
    for dataset_name in _DATASETS:
        dataset = _DATASETS[dataset_name]
        if (registration := dataset.registration) is not None:
            source_dataset = _DATASETS[registration.source_dataset]
            target_dataset = _DATASETS[registration.target_dataset]
            transform = build_transform(
                translation=PhysicalCoordinate(
                    x=registration.translation[0] * target_dataset.data.voxel_size_um,
                    y=registration.translation[1] * target_dataset.data.voxel_size_um,
                    z=registration.translation[2] * target_dataset.data.voxel_size_um,
                ),
                rotation_deg=registration.rotation,
                scale=registration.scale
                * target_dataset.data.voxel_size_um
                / source_dataset.data.voxel_size_um,
            )
            Inventory.add_registration(
                source_dataset=_DATASETS[registration.source_dataset],
                target_dataset=_DATASETS[registration.target_dataset],
                transform=transform,
            )


_populate_registrations_from_metadata(Inventory)



import re

import pytest

from hoa_tools.dataset import _META_DIR, Dataset, change_metadata_directory, get_dataset


@pytest.fixture
def dataset() -> Dataset:
    return get_dataset("LADAF-2020-27_spleen_complete-organ_25.08um_bm05")


def test_dataset_properties() -> None:
    name = "LADAF-2020-27_spleen_complete-organ_25.08um_bm05"
    dataset = get_dataset(name)

    assert dataset.name == name
    assert dataset.is_full_organ
    assert (
        str(dataset) == "Dataset(name=LADAF-2020-27_spleen_complete-organ_25.08um_bm05)"
    )


def test_child_datasets() -> None:
    whole_spleen = get_dataset("LADAF-2020-27_spleen_complete-organ_25.08um_bm05")
    child_datasets = whole_spleen.get_children()
    assert len(child_datasets) == 2

    assert sorted([p.name for p in child_datasets]) == sorted(
        [
            "LADAF-2020-27_spleen_central-column_6.05um_bm05",
            "LADAF-2020-27_spleen_central-column_1.29um_bm05",
        ]
    )


def test_parent_datasets() -> None:
    zoom = get_dataset("LADAF-2020-31_kidney_lateral-transect_2.5um_bm05")
    parent_datasets = zoom.get_parents()
    assert len(parent_datasets) == 1

    assert [p.name for p in parent_datasets] == [
        "LADAF-2020-31_kidney_complete-organ_25.0um_bm05"
    ]


@pytest.mark.vcr
def test_remote_array(dataset: Dataset) -> None:
    data_array = dataset.data_array(downsample_level=2)
    assert data_array.shape == (475, 730, 538)


@pytest.mark.vcr
def test_remote_array_zarr() -> None:
    dataset = get_dataset("A186_lung_right_complete-organ_24.132um_bm18")
    assert dataset._remote_fmt == "zarr"  # noqa: SLF001
    data_array = dataset.data_array(downsample_level=2)
    assert data_array.shape == (2391, 2077, 2077)


def test_invalid_level(dataset: Dataset) -> None:
    with pytest.raises(ValueError, match=re.escape("level must be >= 0")):
        dataset.data_array(downsample_level=-1)  # type: ignore[arg-type]


def test_update_datasets() -> None:
    # Not the best test - this updates the directory to the same previous one
    # But it's at least a smoke test...
    change_metadata_directory(_META_DIR)



# # Improved registrations
#
# Although datasets in the Human Organ Atlas come with a rough registration, sometimes it's useful to do a more accurate registration on a smaller volume of interest of data.
# This can help quantitative comparisons where an accurate registration is important.

# +
import matplotlib.pyplot as plt

import hoa_tools.dataset
import hoa_tools.registration
import hoa_tools.voi

# -

# Lets start by getting an overview dataset, a zoom dataset, and defining a VOI in the zoom dataset.

overview_dataset = hoa_tools.dataset.get_dataset(
    "S-20-29_brain_complete-organ_25.33um_bm05"
)
zoom_dataset = hoa_tools.dataset.get_dataset("S-20-29_brain_VOI-04_6.5um_bm05")
zoom_voi = hoa_tools.voi.VOI(
    dataset=zoom_dataset,
    downsample_level=0,
    lower_corner={"x": 3434, "y": 2060, "z": 2656},
    size={"x": 256, "y": 256, "z": 128},
)


# Get the data in the VOI

# +
zoom_array = zoom_voi.get_data_array()

overview_voi = zoom_voi.transform_to(overview_dataset)
overview_array = overview_voi.get_data_array()
# -

# Plot the data

# +
fig = plt.figure()
ax = zoom_array.isel(z=0).plot(cmap="Grays_r")
ax.axes.set_aspect("equal")
ax.axes.set_title("Zoom dataset")

fig = plt.figure()
ax = overview_array.isel(z=0).plot(cmap="Grays_r")
ax.axes.set_aspect("equal")
ax.axes.set_title("Overview dataset")
# -

# These plots show that the two datasets are defined on different grids.
# In order to do a quanititive comparison, we need to resample one dataset on to the grid of the other datsaet.
# Here we resample the overview VOI on to the grid of the zoom VOI:

import SimpleITK as sitk

resampled_overview = overview_voi.get_data_array_on_voi(
    zoom_voi, interpolator=sitk.sitkNearestNeighbor
)

# Now they are resampled, lets plot the middle slice of each dataset as a simple qualititive comparison

# +
yslice = 70
zslice = 64

fig = plt.figure()
im = zoom_array.isel(z=zslice).plot(cmap="Grays_r")
im.axes.set_aspect("equal")
im.axes.set_title("Zoom dataset")
im.axes.axhline(zoom_array.coords["y"][yslice], color="tab:blue")

fig = plt.figure()
im = resampled_overview.isel(z=zslice).plot(cmap="Grays_r")
im.axes.set_aspect("equal")
im.axes.set_title("Overview dataset\n(resampled on zoom dataset grid)")
im.axes.axhline(resampled_overview.coords["y"][yslice], color="tab:red")

fig, ax = plt.subplots()

zoom_array.isel(z=zslice, y=yslice).plot(ax=ax, color="tab:blue")

ax2 = ax.twinx()
resampled_overview.isel(z=zslice, y=yslice).plot(ax=ax2, color="tab:red")
ax2.set_title("")

ax.set_title("Zoom (blue) and resampled overview (red) comparison")
# -

# ## A more accurate registration
#
# Although it's possible to do a 1:1 comparison after resampling, the images aren't aligned very well.
# In this section we'll step through doing a better registration, using the registration from hoa-tools' registration inventory as a starting point.
#
# To do this we'll use SimpleITK.
# The example here largely follows the [registration introduction](https://insightsoftwareconsortium.github.io/SimpleITK-Notebooks/Python_html/60_Registration_Introduction.html) in the [SimpleITK tutorial notebooks](https://insightsoftwareconsortium.github.io/SimpleITK-Notebooks/).
#
# To start, we'll get both VOIs as SimpleITK images.

zoom_image = zoom_voi.get_sitk_image()
overview_image = overview_voi.get_sitk_image()

# Now we'll get the initial transform from the hoa-tools registration inventory, and then update it so the cnetre of rotation of the transform is in the middle of the zoom image.

initial_transform = hoa_tools.registration.Inventory.get_registration(
    source_dataset=zoom_voi.dataset, target_dataset=overview_voi.dataset
)
initial_transform = sitk.CenteredTransformInitializer(
    zoom_image, overview_image, initial_transform
)

# Now we'll set up the registration and run it.

# +
registration_method = sitk.ImageRegistrationMethod()

registration_method.SetMetricAsMattesMutualInformation(numberOfHistogramBins=50)
registration_method.SetMetricSamplingStrategy(registration_method.RANDOM)
registration_method.SetMetricSamplingPercentage(0.1)

registration_method.SetInterpolator(sitk.sitkLinear)

registration_method.SetOptimizerAsGradientDescent(
    learningRate=1.0,
    numberOfIterations=100,
    convergenceMinimumValue=1e-6,
    convergenceWindowSize=10,
)
registration_method.SetOptimizerScalesFromPhysicalShift()

registration_method.SetInitialTransform(initial_transform, inPlace=False)
final_transform = registration_method.Execute(
    sitk.Cast(zoom_image, sitk.sitkFloat32), sitk.Cast(overview_image, sitk.sitkFloat32)
)
# -

# Using the new transform, resample the overview VOI to the zoom VOI.

resampled_overview = overview_voi.get_data_array_on_voi(
    zoom_voi, interpolator=sitk.sitkLinear, transform=final_transform.GetInverse()
)

# And finally, plot as before.
# Hopefully you can see that the data is aligned much better now, and we can do a more meaningful comparison.

# +
yslice = 180
zslice = 64

fig = plt.figure()
im = zoom_array.isel(z=zslice).plot(cmap="Grays_r")
im.axes.set_aspect("equal")
im.axes.set_title("Zoom dataset")
im.axes.axhline(zoom_array.coords["y"][yslice], color="tab:blue")

fig = plt.figure()
im = resampled_overview.isel(z=zslice).plot(cmap="Grays_r")
im.axes.set_aspect("equal")
im.axes.set_title("Overview dataset\n(resampled on zoom dataset grid)")
im.axes.axhline(resampled_overview.coords["y"][yslice], color="tab:red")

fig, ax = plt.subplots()

zoom_array.isel(z=zslice, y=yslice).plot(ax=ax, color="tab:blue")

ax2 = ax.twinx()
resampled_overview.isel(z=zslice, y=yslice).plot(ax=ax2, color="tab:red")
ax2.set_title("")

ax.set_title("Zoom (blue) and resampled overview (red) comparison")



import pandas as pd

import hoa_tools.inventory


def test_load_inventory() -> None:
    df = hoa_tools.inventory.load_inventory()
    assert isinstance(df, pd.DataFrame)
    assert list(df.columns) == [
        "donor",
        "organ",
        "organ_context",
        "voi",
        "voxel_size_um",
    ]



# # Volumes of interest
#
# In this tutorial we will look at how to represent volumes of
# interest in datasets.

# +
import matplotlib.pyplot as plt

import hoa_tools.dataset
import hoa_tools.voi

# -

# First we'll get an overview dataset, of the brain of donor S-20-29.

overview_dataset = hoa_tools.dataset.get_dataset(
    "S-20-29_brain_complete-organ_25.33um_bm05"
)

# Now we'll define a volume of interest (VOI) object. This represents a cuboid shape volume within the full dataset.

voi = hoa_tools.voi.VOI(
    dataset=overview_dataset,
    downsample_level=4,
    lower_corner={"x": 100, "y": 50, "z": 23},
    size={"x": 40, "y": 20, "z": 1},
)

# Once we have a VOI object, there's a number of useful things we can do.
# Lets get the data array for the VOI and plot it:

data_array = voi.get_data_array()
data_array

plt.figure()
ax = data_array.plot(cmap="Grays_r")
ax.axes.set_aspect("equal")
ax.axes.set_title("Downsample level = 4")


# ## Transforming between dowsample levels
#
# VOIs can be transformed to the different resolution levels of the same dataset.
# Lets get a higher resolution version of the data we plotted above:

high_res_voi = voi.change_downsample_level(new_downsample_level=2)

# Because we've increased the resolution (equivalently, decreased the downsampling level)
# the VOI is now bigger.

high_res_data_array = high_res_voi.get_data_array()

# +
fig, axs = plt.subplots(nrows=2)
data_array.isel(z=0).plot(cmap="Grays_r", ax=axs[0])
axs[0].set_aspect("equal")
axs[0].set_title("Downsample level = 4")

high_res_data_array.isel(z=0).plot(cmap="Grays_r", ax=axs[1])
axs[1].set_aspect("equal")
axs[1].set_title("Downsample level = 2")

fig.tight_layout()
# -

# ## Transforming between registered datasets
#
# Now we'll step through transforming a VOI between two datasets that have been registered to each other.
# We'll start by getting one of the children of the overview dataset used above.

children = overview_dataset.get_children()
child = children[8]
child.name

child_array = child.data_array(downsample_level=0)
child_array

# Now we'll select a smaller volume of interest from the child array, and plot it

child_voi = hoa_tools.voi.VOI(
    dataset=child,
    downsample_level=0,
    lower_corner={"x": 3434, "y": 2060, "z": 2656},
    size={"x": 256, "y": 256, "z": 128},
)
child_array = child_voi.get_data_array()

plt.figure()
ax = child_array.isel(z=0).plot(cmap="Grays_r")
ax.axes.set_aspect("equal")
ax.axes.set_title("Zoom dataset slice")


# Now lets transform the VOI to the overview dataset

overview_voi = child_voi.transform_to(overview_dataset)
overview_voi

# Finally, lets plot the overview and child datasets together. We can see that the same structures are visible in both, so the registration is approximately accurate, and all the transformations worked.
#
# In the next tutorial section we'll see how to do a more accurate registration of just these small sub-volumes of data in order to do quantitative comparisons.

overview_array = overview_voi.get_data_array()
overview_array

# +
fig, axs = plt.subplots(nrows=2, figsize=(6, 9))

ax = axs[0]
child_array.isel(z=0).plot(cmap="Grays_r", ax=ax)
ax.set_aspect("equal")
ax.set_title("Zoom dataset")

ax = axs[1]
overview_array.isel(z=0).plot(cmap="Grays_r", ax=ax)
ax.set_aspect("equal")
ax.set_title("Overview dataset")
fig.tight_layout()



from typing import Any

import zarr.n5
from zarr.storage import FSStore


# Override __init__ to prevent the deprecation warning from showing up
class N5FSStore(zarr.n5.N5FSStore):  # type: ignore[misc]
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        FSStore.__init__(self, *args, **kwargs)
